@@ -83,6 +83,14 @@ class XPathExpr(object):
 
 split_at_single_quotes = re.compile("('+)").split
 
+# The spec is actually more permissive than that, but don’t bother.
+# This is just for the fast path.
+# http://www.w3.org/TR/REC-xml/#NT-NameStartChar
+is_safe_name = re.compile('^[a-zA-Z_][a-zA-Z0-9_.-]*$').match
+
+# Test that the string is not empty and does not contain whitespace
+is_non_whitespace = re.compile(r'^[^ \t\r\n\f]+$').match
+
 
 #### Translation
 
@@ -94,6 +102,19 @@ class GenericTranslator(object):
     of element names and attribute names.
 
     """
+    
+    ####
+    ####  HERE BE DRAGONS
+    ####
+    ####  You are welcome to hook into this to change some behavior,
+    ####  but do so at your own risks.
+    ####  Until is has recieved a lot more work and review,
+    ####  I reserve the right to change this API in backward-incompatible ways
+    ####  with any minor version of cssselect.
+    ####  See https://github.com/SimonSapin/cssselect/pull/22
+    ####  -- Simon Sapin.
+    ####
+
     combinator_mapping = {
         ' ': 'descendant',
         '>': 'child',
@@ -138,6 +159,9 @@ class GenericTranslator(object):
     lower_case_attribute_names = False
     lower_case_attribute_values = False
 
+    # class used to represent and xpath expression
+    xpathexpr_cls = XPathExpr
+
     def css_to_xpath(self, css, prefix='descendant-or-self::'):
         """Translate a *group of selectors* to XPath.
 
@@ -181,7 +205,9 @@ class GenericTranslator(object):
         tree = getattr(selector, 'parsed_tree', None)
         if not tree:
             raise TypeError('Expected a parsed selector, got %r' % (selector,))
-        return (prefix or '') + _unicode(self.xpath(tree))
+        xpath = self.xpath(tree)
+        assert isinstance(xpath, self.xpathexpr_cls)  # help debug a missing 'return'
+        return (prefix or '') + _unicode(xpath)
 
     @staticmethod
     def xpath_literal(s):
@@ -224,7 +250,9 @@ class GenericTranslator(object):
 
     def xpath_function(self, function):
         """Translate a functional pseudo-class."""
-        method = 'xpath_%s_function' % function.name.replace('-', '_').lower()
+        method = 'xpath_%s_function' % function.name.replace('-', '_')
+        # getattr() with a non-ASCII name fails on Python 2.x
+        method = method.encode('ascii', 'replace').decode('ascii')
         method = getattr(self, method, None)
         if not method:
             raise ExpressionError(
@@ -233,7 +261,9 @@ class GenericTranslator(object):
 
     def xpath_pseudo(self, pseudo):
         """Translate a pseudo-class."""
-        method = 'xpath_%s_pseudo' % pseudo.ident.replace('-', '_').lower()
+        method = 'xpath_%s_pseudo' % pseudo.ident.replace('-', '_')
+        # getattr() with a non-ASCII name fails on Python 2.x
+        method = method.encode('ascii', 'replace').decode('ascii')
         method = getattr(self, method, None)
         if not method:
             # TODO: better error message for pseudo-elements?
@@ -250,15 +280,19 @@ class GenericTranslator(object):
             name = selector.attrib.lower()
         else:
             name = selector.attrib
-        if selector.namespace == '*':
-            name = '@' + name
+        safe = is_safe_name(name)
+        if selector.namespace:
+            name = '%s:%s' % (selector.namespace, name)
+            safe = safe and is_safe_name(selector.namespace)
+        if safe:
+            attrib = '@' + name
         else:
-            name = '@%s:%s' % (selector.namespace, name)
+            attrib = 'attribute::*[name() = %s]' % self.xpath_literal(name)
         if self.lower_case_attribute_values:
             value = selector.value.lower()
         else:
             value = selector.value
-        return method(self.xpath(selector.selector), name, value)
+        return method(self.xpath(selector.selector), attrib, value)
 
     def xpath_class(self, class_selector):
         """Translate a class selector."""
@@ -274,15 +308,23 @@ class GenericTranslator(object):
 
     def xpath_element(self, selector):
         """Translate a type or universal selector."""
-        if self.lower_case_element_names:
-            element = selector.element.lower()
+        element = selector.element
+        if not element:
+            element = '*'
+            safe = True
         else:
-            element = selector.element
-        if selector.namespace != '*':
+            safe = is_safe_name(element)
+            if self.lower_case_element_names:
+                element = element.lower()
+        if selector.namespace:
             # Namespace prefixes are case-sensitive.
             # http://www.w3.org/TR/css3-namespace/#prefixes
             element = '%s:%s' % (selector.namespace, element)
-        return XPathExpr(element=element)
+            safe = safe and is_safe_name(selector.namespace)
+        xpath = self.xpathexpr_cls(element=element)
+        if not safe:
+            xpath.add_name_test()
+        return xpath
 
 
     # CombinedSelector: dispatch by combinator
@@ -367,18 +409,24 @@ class GenericTranslator(object):
                                              add_name_test=False)
 
     def xpath_contains_function(self, xpath, function):
-        return xpath.add_condition('contains(string(.), %s)'
-                            % self.xpath_literal(function.arguments))
+        # Defined there, removed in later drafts:
+        # http://www.w3.org/TR/2001/CR-css3-selectors-20011113/#content-selectors
+        if function.argument_types() not in (['STRING'], ['IDENT']):
+            raise ExpressionError(
+                "Expected a single string or ident for :contains(), got %r"
+                % function.arguments)
+        value = function.arguments[0].value
+        return xpath.add_condition(
+            'contains(string(.), %s)' % self.xpath_literal(value))
 
     def xpath_lang_function(self, xpath, function):
+        if function.argument_types() not in (['STRING'], ['IDENT']):
+            raise ExpressionError(
+                "Expected a single string or ident for :lang(), got %r"
+                % function.arguments)
+        value = function.arguments[0].value
         return xpath.add_condition(
-            "ancestor-or-self::*[@lang][1][starts-with(concat("
-                # XPath 1.0 has no lower-case function...
-                "translate(@%s, 'ABCDEFGHIJKLMNOPQRSTUVWXYZ', "
-                               "'abcdefghijklmnopqrstuvwxyz'), "
-                "'-'), %s)]"
-            % (self.lang_attribute, self.xpath_literal(
-                function.arguments.lower() + '-')))
+            "lang(%s)" % (self.xpath_literal(value)))
 
 
     # Pseudo: dispatch by pseudo-class name
@@ -422,7 +470,7 @@ class GenericTranslator(object):
         return xpath.add_condition('last() = 1')
 
     def xpath_empty_pseudo(self, xpath):
-        return xpath.add_condition("not(*) and not(normalize-space())")
+        return xpath.add_condition("not(*) and not(string-length())")
 
     def pseudo_never_matches(self, xpath):
         """Common implementation for pseudo-classes that never match."""
@@ -460,9 +508,12 @@ class GenericTranslator(object):
         return xpath
 
     def xpath_attrib_includes(self, xpath, name, value):
-        xpath.add_condition(
-            "%s and contains(concat(' ', normalize-space(%s), ' '), %s)"
-            % (name, name, self.xpath_literal(' '+value+' ')))
+        if is_non_whitespace(value):
+            xpath.add_condition(
+                "%s and contains(concat(' ', normalize-space(%s), ' '), %s)"
+                % (name, name, self.xpath_literal(' '+value+' ')))
+        else:
+            xpath.add_condition('0')
         return xpath
 
     def xpath_attrib_dashmatch(self, xpath, name, value):
@@ -474,19 +525,31 @@ class GenericTranslator(object):
         return xpath
 
     def xpath_attrib_prefixmatch(self, xpath, name, value):
-        return xpath.add_condition('%s and starts-with(%s, %s)' % (
-            name, name, self.xpath_literal(value)))
+        if value:
+            xpath.add_condition('%s and starts-with(%s, %s)' % (
+                name, name, self.xpath_literal(value)))
+        else:
+            xpath.add_condition('0')
+        return xpath
 
     def xpath_attrib_suffixmatch(self, xpath, name, value):
-        # Oddly there is a starts-with in XPath 1.0, but not ends-with
-        return xpath.add_condition(
-            '%s and substring(%s, string-length(%s)-%s) = %s'
-            % (name, name, name, len(value)-1, self.xpath_literal(value)))
+        if value:
+            # Oddly there is a starts-with in XPath 1.0, but not ends-with
+            xpath.add_condition(
+                '%s and substring(%s, string-length(%s)-%s) = %s'
+                % (name, name, name, len(value)-1, self.xpath_literal(value)))
+        else:
+            xpath.add_condition('0')
+        return xpath
 
     def xpath_attrib_substringmatch(self, xpath, name, value):
-        # Attribute selectors are case sensitive
-        return xpath.add_condition('%s and contains(%s, %s)' % (
-            name, name, self.xpath_literal(value)))
+        if value:
+            # Attribute selectors are case sensitive
+            xpath.add_condition('%s and contains(%s, %s)' % (
+                name, name, self.xpath_literal(value)))
+        else:
+            xpath.add_condition('0')
+        return xpath
 
 
 class HTMLTranslator(GenericTranslator):
@@ -523,6 +586,20 @@ class HTMLTranslator(GenericTranslator):
                 "and (name(.) = 'input' or name(.) = 'command')"
                 "and (@type = 'checkbox' or @type = 'radio'))")
 
+    def xpath_lang_function(self, xpath, function):
+        if function.argument_types() not in (['STRING'], ['IDENT']):
+            raise ExpressionError(
+                "Expected a single string or ident for :lang(), got %r"
+                % function.arguments)
+        value = function.arguments[0].value
+        return xpath.add_condition(
+            "ancestor-or-self::*[@lang][1][starts-with(concat("
+                # XPath 1.0 has no lower-case function...
+                "translate(@%s, 'ABCDEFGHIJKLMNOPQRSTUVWXYZ', "
+                               "'abcdefghijklmnopqrstuvwxyz'), "
+                "'-'), %s)]"
+            % (self.lang_attribute, self.xpath_literal(value.lower() + '-')))
+
     def xpath_link_pseudo(self, xpath):
         return xpath.add_condition("@href and "
             "(name(.) = 'a' or name(.) = 'link' or name(.) = 'area')")
